from django import template
from django.apps import apps
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ObjectDoesNotExist
from django.shortcuts import get_object_or_404

from apis_core.core.templatetags.core import get_model_fields
from apis_core.generic.abc import GenericModel
from apis_core.generic.helpers import template_names_via_mro

register = template.Library()


@register.filter
def contenttype(model):
    return ContentType.objects.get_for_model(model)


@register.simple_tag
def modeldict(instance, fields=None, exclude=None):
    data = {}
    for f in get_model_fields(instance):
        if not getattr(f, "editable", False):
            continue
        if fields is not None and f.name not in fields:
            continue
        if exclude and f.name in exclude:
            continue
        field = instance._meta.get_field(f.name)
        try:
            data[field] = getattr(instance, field.name)
        except ObjectDoesNotExist as e:
            data[field] = f"{field.value_from_object(instance)} ({e})"
        if fn := getattr(instance, f"get_{field.name}_display", False):
            data[field] = fn()
        if getattr(field, "m2m_field_name", False):
            values = getattr(instance, field.name).all()
            data[field] = ", ".join([str(value) for value in values])
    return data


@register.simple_tag
def contenttypes(app_labels=None):
    if app_labels:
        app_labels = app_labels.split(",")
        return ContentType.objects.filter(app_label__in=app_labels)
    return ContentType.objects.all()


def is_genericmodel(content_type: ContentType):
    model_class = content_type.model_class()
    return model_class is not None and issubclass(model_class, GenericModel)


@register.simple_tag
def genericmodel_content_types():
    """
    Retrieve all models which inherit from GenericModel class
    and return their ContentType.
    """
    genericmodels = list(
        filter(
            lambda content_type: is_genericmodel(content_type),
            ContentType.objects.all(),
        )
    )
    return genericmodels


@register.simple_tag
def pure_genericmodel_content_types():
    """
    Retrieve all models which inherit from GenericModel class
    but are not Collections, Entities, Relations or History models
    """
    parents = []
    if apps.is_installed("apis_core.collections"):
        collections = apps.get_app_config("collections")
        parents.append(collections.models_module.SkosCollection)
        parents.append(collections.models_module.SkosCollectionContentObject)
    if apps.is_installed("apis_core.relations"):
        relations = apps.get_app_config("relations")
        parents.append(relations.models_module.Relation)
    if apps.is_installed("apis_core.history"):
        history = apps.get_app_config("history")
        parents.append(history.models_module.APISHistoryTableBase)
    if apps.is_installed("apis_core.apis_entities"):
        entities = apps.get_app_config("apis_entities")
        parents.append(entities.models_module.AbstractEntity)
    genericmodels = [
        ct
        for ct in set(genericmodel_content_types())
        if not issubclass(ct.model_class(), tuple(parents))
    ]
    return genericmodels


@register.filter
def get_attribute(obj, attribute):
    return getattr(obj, attribute, None)


@register.filter
def content_type_count(content_type):
    """
    Return the number of objects having a specific content type
    """
    return content_type.model_class().objects.count()


@register.simple_tag
def template_list(obj, suffix):
    return template_names_via_mro(type(obj), suffix)


@register.simple_tag(takes_context=True)
def any_view_permission(context, content_types):
    user = context.request.user
    return any(
        [user.has_perm(ct.model_class().get_view_permission()) for ct in content_types]
    )


@register.simple_tag
def content_types_by_natural_keys(natural_keys: tuple() = ()) -> [ContentType]:
    """
    Convert a list of natural keys to a list of ContentType models
    If any of the natural keys does not refer to an existing model, raise a 404
    """
    content_types = []
    for key in natural_keys:
        app_label, model = key.split(".")
        content_type = get_object_or_404(ContentType, app_label=app_label, model=model)
        content_types.append(content_type)
    return content_types


@register.simple_tag
def natural_keys_by_content_types(content_types: tuple() = ()) -> [str]:
    """
    Convert a list of ContentType models to their natural key
    """
    natural_keys = []
    for content_type in content_types:
        natural_keys.append(content_type.app_label + "." + content_type.model)
    return natural_keys


@register.filter
def split(string: str = "", delimiter=",") -> [str]:
    """
    Split a string by a specific delimiter and also strip the string of
    leading and trailing whitespaces.
    """
    return map(str.strip, string.split(delimiter))


@register.simple_tag
def model_field_template_lookup_list(model, field, suffix="") -> [str]:
    """
    generate a template path based on the modelname, the fieldname and
    the suffix. return a list with this template path and a fallback
    path.
    """
    content_type = ContentType.objects.get_for_model(model)
    path = f"{content_type.app_label}/partials/{content_type.model}_{field.name}_{suffix}.html"
    return [path, f"generic/partials/default_model_field_{suffix}.html"]



import pathlib
import secrets
from datetime import datetime

from django.contrib.auth.models import User
from django.core.management.base import BaseCommand


class Command(BaseCommand):
    help = "Set up APIS test instance"

    def handle(self, *app_labels, **options):
        password = secrets.token_urlsafe(8)
        User.objects.create_superuser(
            username="admin", email="admin@example.org", password=password
        )
        pathlib.Path("/tmp/password.txt").write_text(password)
        pathlib.Path("/tmp/startup.txt").write_text(str(datetime.utcnow()))



from django.forms.fields import ChoiceField
from django.utils.translation import gettext_lazy as _

from apis_core.apis_entities.fields import PlaceLookupField
from apis_core.apis_entities.utils import get_feature_codes
from apis_core.generic.forms import GenericModelForm


class E53_PlaceForm(GenericModelForm):
    place = PlaceLookupField(
        required=False,
        label=_("Search places"),
        help_text=_("Click on a result to fill the form"),
    )
    feature_code = ChoiceField(choices=get_feature_codes, required=False)

    field_order = ["place"]



# Generated by Django 3.1.14 on 2022-02-18 15:30

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("apis_metainfo", "0002_auto_20220201_1241"),
    ]

    operations = [
        migrations.AlterField(
            model_name="text",
            name="text",
            field=models.TextField(blank=True, default=""),
            preserve_default=False,
        ),
    ]



from apis_core.generic.tables import ActionColumn, GenericTable


class DuplicateColumn(ActionColumn):
    """
    A column showing a view button
    """

    template_name = "columns/duplicate.html"
    permission = "create"


class AbstractEntityTable(GenericTable):
    noduplicate = DuplicateColumn()

    class Meta(GenericTable.Meta):
        sequence = ("...", "view", "edit", "delete", "noduplicate")



# Generated by Django 5.2.4 on 2025-08-05 09:27

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("apis_metainfo", "0016_remove_rootobject_self_contenttype"),
    ]

    operations = [
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.DeleteModel(
                    name="Uri",
                ),
            ],
            database_operations=[
                migrations.AlterModelTable(
                    name="Uri",
                    table="uris_uri",
                ),
            ],
        )
    ]



from django_filters.filterset import filterset_factory
from django_filters.rest_framework import DjangoFilterBackend

from .filtersets import GenericFilterSet
from .helpers import first_member_match, module_paths


class GenericFilterBackend(DjangoFilterBackend):
    def get_filterset_class(self, view, queryset=None):
        filterset_api_class_modules = module_paths(
            view.model, path="filtersets", suffix="ApiFilterSet"
        )
        filterset_class_modules = module_paths(
            view.model, path="filtersets", suffix="FilterSet"
        )
        filterset_class = first_member_match(
            filterset_api_class_modules + filterset_class_modules, GenericFilterSet
        )
        return filterset_factory(view.model, filterset_class)



# Generated by Django 4.2.10 on 2024-04-11 09:26

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("apis_metainfo", "0012_remove_rootobject_deprecated_name"),
    ]

    operations = [
        migrations.DeleteModel(
            name="Collection",
        ),
    ]



# Generated by Django 5.1.4 on 2025-02-19 17:43

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("apis_metainfo", "0015_uri_content_type_uri_object_id"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="rootobject",
            name="self_contenttype",
        ),
    ]



import django_tables2 as tables

from apis_core.generic.tables import GenericTable

from .models import Uri


class UriTable(GenericTable):
    entity = tables.TemplateColumn(
        "<a href='{{ record.content_object.get_absolute_url }}'>{{ record.content_object }}</a>",
        orderable=False,
        verbose_name="related Entity",
    )
    content_type = tables.TemplateColumn(
        "{{ record.content_type.model }}",
        verbose_name="Entity Type",
    )

    class Meta(GenericTable.Meta):
        model = Uri
        fields = ["id", "uri", "entity", "content_type"]
        exclude = ("desc",)



from django.contrib.auth.models import User
from django.test import TestCase
from faker import Faker

from apis_core.collections.models import SkosCollection, SkosCollectionContentObject

fake = Faker()


class SkosCollectionTestCase(TestCase):
    def test_name(self):
        name = fake.name()
        skc = SkosCollection.objects.create(name=name)
        self.assertEqual(str(skc), name)


class SkosCollectionContentObjectTestCase(TestCase):
    def setUp(self):
        self.collection_name = fake.color_name()
        self.collection = SkosCollection.objects.create(name=self.collection_name)

    def test_collection(self):
        user = User.objects.create(username=fake.name())
        scco = SkosCollectionContentObject(
            collection=self.collection, content_object=user
        )
        self.assertEqual(str(scco), f"{user} -> {self.collection_name}")



# Generated by Django 4.1.10 on 2023-09-13 12:17

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("apis_entities", "0002_remove_tempentityclass_text"),
        ("apis_metainfo", "0005_delete_uricandidate"),
    ]

    operations = [
        migrations.DeleteModel(
            name="Text",
        ),
    ]



import logging
from pathlib import Path

from django.conf import settings
from django.shortcuts import render

logger = logging.getLogger(__name__)


class MaintenanceMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        maintenance_file = getattr(
            settings, "APIS_MAINTENANCE_FILE", "apis_maintenance"
        )
        if Path(maintenance_file).exists():
            logger.warning("Site is running in maintenance mode")
            if hasattr(request, "user"):
                if request.user.is_superuser:
                    return self.get_response(request)
            return render(request, "maintenance.html")
        return self.get_response(request)



from django.forms import Form
from django.views.generic.detail import BaseDetailView, DetailView
from django.views.generic.edit import FormView
from django_tables2 import SingleTableMixin
from django_tables2.tables import table_factory

from apis_core.generic.helpers import first_member_match, module_paths
from apis_core.generic.views import GenericModelMixin

from .tables import HistoryGenericTable


class HistoryView(GenericModelMixin, SingleTableMixin, DetailView):
    template_name = "history/history.html"

    def get_table_class(self):
        table_modules = module_paths(self.model, path="tables", suffix="HistoryTable")
        table_class = first_member_match(table_modules, HistoryGenericTable)
        return table_factory(self.model, table_class)

    def get_table_data(self):
        return self.get_object().get_history_data()


class HistoryReset(GenericModelMixin, BaseDetailView, FormView):
    form_class = Form
    template_name = "history/reset.html"

    def form_valid(self, form):
        self.get_object().instance.save()
        return super().form_valid(form)

    def get_success_url(self):
        return self.get_object().instance.get_history_url()



from django.views.generic.base import TemplateView

from apis_core.documentation.utils import Datamodel


class Documentation(TemplateView):
    template_name = "documentation.html"

    def get_context_data(self):
        ctx = super().get_context_data()
        ctx["datamodel"] = Datamodel()
        return ctx



from django.urls import path

from apis_core.apis_entities.api_views import GetEntityGeneric, ListEntityGeneric

# from .views import ReversionCompareView TODO: add again when import is fixed
from apis_core.apis_entities.views import E53_PlaceMap, EntitiesAutocomplete

api_routes = [
    path("entities/", ListEntityGeneric.as_view()),
    path(
        "entity/<int:pk>/",
        GetEntityGeneric.as_view(),
        name="GetEntityGeneric",
    ),
]


app_name = "apis_entities"

urlpatterns = [
    path("autocomplete/", EntitiesAutocomplete.as_view(), name="autocomplete"),
    path("<contenttype:contenttype>/map", E53_PlaceMap.as_view(), name="e53_place_map"),
]



# Generated by Django 5.1.7 on 2025-04-16 10:06

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("relations", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="relation",
            name="obj_object_id",
            field=models.PositiveIntegerField(null=True),
        ),
        migrations.AlterField(
            model_name="relation",
            name="subj_object_id",
            field=models.PositiveIntegerField(null=True),
        ),
    ]



from django import forms
from django.urls import reverse

from apis_core.core.fields import ApisListSelect2
from apis_core.relations.utils import relation_content_types


def list_relation_choices():
    return [(rel.id, rel.model_class().name()) for rel in relation_content_types()]


class RelationMultiWidget(forms.MultiWidget):
    template_name = "relations/relation_multiwidget.html"
    use_fieldset = False

    def __init__(self, attrs=None):
        relation_choices = [(None, "---")] + list_relation_choices()
        widgets = [
            forms.widgets.Select(choices=relation_choices),
            ApisListSelect2(url=reverse("apis_core:apis_entities:autocomplete")),
        ]
        super().__init__(widgets)

    def decompress(self, value):
        return [value, value]


class RelationField(forms.MultiValueField):
    widget = RelationMultiWidget()

    def __init__(self, *args, **kwargs):
        fields = (forms.ChoiceField(choices=list_relation_choices()), forms.CharField())
        super().__init__(fields=fields, *args, **kwargs)

    def compress(self, data_list):
        return data_list



import logging

from rdflib import Graph
from rest_framework import renderers
from rest_framework.exceptions import APIException

logger = logging.getLogger(__name__)


class GenericRDFBaseRenderer(renderers.BaseRenderer):
    """
    Base class to render RDF graphs to various formats.
    This renderer expects the serialized data to either be a rdflib grap **or**
    to contain a list of rdflib graphs. If it works with a list of graphs, those
    are combined to one graph.
    This graph is then serialized and the result is returned. The serialization
    format can be set using the `rdflib_format` attribute. If this is not set, the
    `format` attribute of the renderer is used as serialization format (this is the
    format as it is used by the Django Rest Framework for content negotiation.
    """

    format = "ttl"
    rdflib_format = None

    def render(self, data, accepted_media_type=None, renderer_context=None):
        result = Graph()

        match data:
            case {"results": results, **rest}:  # noqa: F841
                # Handle case where data is a dict with multiple graphs
                for graph in results:
                    if isinstance(graph, Graph):
                        # Merge triples
                        for triple in graph:
                            result.add(triple)
                        # Merge namespace bindings
                        for prefix, namespace in graph.namespaces():
                            result.bind(prefix, namespace, override=False)
            case {"detail": detail}:
                raise APIException(detail)
            case Graph():
                # Handle case where data is a single graph
                result = data
                # Ensure namespaces are properly bound in the single graph case
                for prefix, namespace in data.namespaces():
                    result.bind(prefix, namespace, override=False)
            case _:
                raise ValueError(
                    "Invalid data format. Expected rdflib Graph or dict with 'results' key containing graphs"
                )
        serialization_format = self.rdflib_format or self.format
        return result.serialize(format=serialization_format)


class GenericRDFTurtleRenderer(GenericRDFBaseRenderer):
    format = "ttl"
    media_type = "text/turtle"
    rdflib_format = "turtle"


class GenericRDFXMLRenderer(GenericRDFBaseRenderer):
    format = "rdf"
    media_type = "application/rdf+xml"
    rdflib_format = "xml"


class GenericRDFN3Renderer(GenericRDFBaseRenderer):
    format = "rdf"
    media_type = "text/n3"
    rdflib_format = "n3"


class CidocTTLRenderer(GenericRDFBaseRenderer):
    format = "cidoc"
    media_type = "text/ttl"
    rdflib_format = "ttl"


class CidocXMLRenderer(GenericRDFBaseRenderer):
    format = "cidoc"
    media_type = "application/rdf+xml"
    rdflib_format = "xml"



import logging

from crispy_forms.helper import FormHelper
from crispy_forms.layout import Submit
from django import forms
from django.apps import apps
from django.contrib.contenttypes.models import ContentType
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from apis_core.core.fields import ApisModelSelect2, ApisModelSelect2Multiple
from apis_core.generic.abc import GenericModel
from apis_core.generic.forms.fields import ModelImportChoiceField

logger = logging.getLogger(__name__)


class GenericImportForm(forms.Form):
    class Meta:
        fields = []

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["url"] = ModelImportChoiceField(
            queryset=self.Meta.model.objects.all()
        )
        ct = ContentType.objects.get_for_model(self.Meta.model)
        url = reverse("apis_core:generic:autocompleteexternalonly", args=[ct])
        self.fields["url"].widget = ApisModelSelect2(
            url, attrs={"data-html": True, "data-tags": 1}
        )
        self.fields["url"].widget.choices = self.fields["url"].choices
        self.helper = FormHelper()
        self.helper.add_input(Submit("submit", _("Submit")))


class GenericFilterSetForm(forms.Form):
    """
    FilterSet form for generic models
    Adds a submit button using the django crispy form helper
    Adds a `columns` selector that lists all the fields from
    the model
    """

    columns_exclude = []

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.helper = FormHelper(self)
        self.helper.form_method = "GET"
        self.helper.add_input(Submit("submit", _("Submit")))

    def clean(self):
        self.cleaned_data = super().clean()
        self.cleaned_data.pop("columns", None)
        return self.cleaned_data


class GenericModelForm(forms.ModelForm):
    """
    Model form for generic models
    Adds a submit button using the django crispy form helper
    and sets the ModelChoiceFields and ModelMultipleChoiceFields
    to use autocomplete replacement fields
    """

    class Meta:
        fields = "__all__"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        try:
            skoscollection = apps.get_model("collections.SkosCollection")
            self.fields["collections"] = forms.ModelMultipleChoiceField(
                required=False,
                queryset=skoscollection.objects.all(),
                label=_("Collections"),
            )
            if instance := kwargs.get("instance"):
                self.fields["collections"].initial = skoscollection.objects.by_instance(
                    instance
                ).values_list("pk", flat=True)
        except LookupError as e:
            logger.debug("Not adding collections to form: %s", e)

        self.helper = FormHelper(self)
        self.helper.add_input(Submit("submit", _("Submit")))

        # override the fields pointing to other models,
        # to make them use the autocomplete widgets
        override_fieldtypes = {
            "ModelMultipleChoiceField": ApisModelSelect2Multiple,
            "ModelChoiceField": ApisModelSelect2,
            "ModelImportChoiceField": ApisModelSelect2,
        }
        for field in self.fields:
            clsname = self.fields[field].__class__.__name__
            if clsname in override_fieldtypes.keys():
                ct = ContentType.objects.get_for_model(
                    self.fields[field]._queryset.model
                )
                if issubclass(ct.model_class(), GenericModel):
                    url = reverse("apis_core:generic:autocomplete", args=[ct])
                    self.fields[field].widget = override_fieldtypes[clsname](
                        url, attrs={"data-html": True}
                    )
                    self.fields[field].widget.choices = self.fields[field].choices

    def save(self, *args, **kwargs):
        instance = super().save(*args, **kwargs)
        try:
            skoscollection = apps.get_model("collections.SkosCollection")
            if "collections" in self.cleaned_data:
                collections = self.cleaned_data.get("collections")
                for collection in skoscollection.objects.exclude(pk__in=collections):
                    collection.remove(instance)
                for collection in skoscollection.objects.filter(pk__in=collections):
                    collection.add(instance)
        except LookupError as e:
            logger.debug("Not creating collections from form: %s", e)
        return instance


class GenericSelectMergeOrEnrichForm(forms.Form):
    def __init__(self, *args, **kwargs):
        if "instance" in kwargs:
            instance = kwargs.pop("instance")
        super().__init__(*args, **kwargs)
        ct = ContentType.objects.get_for_model(instance)
        self.fields["uri"] = forms.ModelChoiceField(
            queryset=ct.model_class().objects.all()
        )
        uri = reverse("apis_core:generic:autocomplete", args=[ct])
        attrs = {"data-html": True, "data-tags": 1}
        self.fields["uri"].widget = ApisModelSelect2(uri, attrs=attrs)
        self.fields["uri"].widget.choices = self.fields["uri"].choices
        self.fields["uri"].label = "Select or paste URI"
        self.helper = FormHelper()
        self.helper.form_method = "GET"
        self.helper.add_input(Submit("submit", _("Submit")))
        self.helper.form_action = instance.get_enrich_url()


class GenericMergeWithForm(forms.Form):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.add_input(Submit("submit", _("Merge")))


class GenericEnrichForm(forms.Form):
    def __init__(self, *args, **kwargs):
        data = kwargs.pop("data", {})
        instance = kwargs.pop("instance", None)
        super().__init__(*args, **kwargs)
        for key, value in data.items():
            update_key = f"update_{key}"
            self.fields[update_key] = forms.BooleanField(
                required=False,
                label=f"Update {key} from {getattr(instance, key)} to {value}",
            )

            self.fields[key] = forms.CharField(initial=value, required=False)
            self.fields[key].widget = self.fields[key].hidden_widget()
        self.helper = FormHelper()
        self.helper.add_input(Submit("submit", _("Submit")))



from django.db.models.signals import ModelSignal

pre_merge_with = ModelSignal(use_caching=True)
post_merge_with = ModelSignal(use_caching=True)
pre_duplicate = ModelSignal(use_caching=True)
post_duplicate = ModelSignal(use_caching=True)



from rest_framework import serializers

from apis_core.generic.serializers import GenericHyperlinkedModelSerializer

from .models import Uri


class UriSerializer(GenericHyperlinkedModelSerializer):
    url = serializers.HyperlinkedIdentityField(
        view_name="apis_core:apis_api:uri-detail", lookup_field="pk"
    )

    class Meta:
        fields = "__all__"
        model = Uri



from rdflib import Literal, URIRef
from rdflib.namespace import GEO, RDF, RDFS
from rest_framework import serializers

from apis_core.generic.serializers import (
    GenericHyperlinkedIdentityField,
    GenericModelCidocSerializer,
)
from apis_core.generic.utils.rdf_namespace import APPELLATION, CRM


class MinimalEntitySerializer(serializers.Serializer):
    uri = GenericHyperlinkedIdentityField(
        view_name="apis_core:generic:genericmodelapi-detail"
    )
    name = serializers.SerializerMethodField(method_name="get_name")

    def get_name(self, object) -> str:
        return str(object)


class E21_PersonCidocSerializer(GenericModelCidocSerializer):
    def to_representation(self, instance):
        g = super().to_representation(instance)

        if hasattr(instance, "forename"):
            forename_uri = URIRef(APPELLATION[f"forename_{instance.id}"])
            g.add(
                (
                    forename_uri,
                    RDF.type,
                    CRM.E33_E41_Linguistic_Appellation,
                )
            )
            g.add(
                (
                    self.appellation_uri,
                    CRM.P106_is_composed_of,
                    forename_uri,
                )
            )
            g.add((forename_uri, RDFS.label, Literal(instance.forename)))

        if hasattr(instance, "surname"):
            surname_uri = URIRef(APPELLATION[f"surname_{instance.id}"])
            g.add(
                (
                    surname_uri,
                    RDF.type,
                    CRM.E33_E41_Linguistic_Appellation,
                )
            )
            g.add(
                (
                    self.appellation_uri,
                    CRM.P106_is_composed_of,
                    surname_uri,
                )
            )
            g.add((surname_uri, RDFS.label, Literal(instance.surname)))

        return g


class E53_PlaceCidocSerializer(GenericModelCidocSerializer):
    def to_representation(self, instance):
        g = super().to_representation(instance)

        if instance.latitude is not None and instance.longitude is not None:
            literal_text = f"Point ( {instance.longitude:+} {instance.latitude:+} )"
            literal = Literal(literal_text, datatype=GEO.wktLiteral)
            g.add((self.instance_uri, CRM.P168_place_is_defined_by, literal))
        return g



import pytest
from django.contrib.auth.models import Permission, User
from django.contrib.contenttypes.models import ContentType
from faker import Faker

from apis_core.generic.tests.models import Dummy, Person
from apis_core.generic.tests.tables import DummyTable
from apis_core.generic.views import List

fake = Faker()

# tuple of route, permission, and expected status_code
routes = [
    ("/generic_tests.person/", "view_person", 200),
    ("/generic_tests.person/1", "view_person", 200),
    ("/generic_tests.person/create", "add_person", 200),
    ("/generic_tests.person/delete/1", "delete_person", 200),
    ("/generic_tests.person/update/1", "change_person", 200),
    ("/generic_tests.person/duplicate/1", "add_person", 302),
    ("/generic_tests.person/autocomplete", "view_person", 200),
    ("/generic_tests.person/import", "add_person", 200),
    ("/generic_tests.person/selectmergeorenrich/1", "add_person", 200),
    ("/generic_tests.person/merge/1/2", "change_person", 200),
    (
        "/generic_tests.person/enrich/1?uri=https://d-nb.info/gnd/119252880",
        "change_person",
        200,
    ),
]
route_parameters = "route,permission,status_code"

unique_first_name = "SomeUniqueFirstName"
unique_last_name = "SomeUniqueLastName"


@pytest.mark.django_db
class TestGeneric(object):
    @pytest.fixture
    def setup(self):
        # Create an admin user
        self.admin = User.objects.create_superuser(
            "admin", fake.email(), fake.password()
        )
        # Create a normal user
        self.user = User.objects.create_user(
            fake.user_name(), fake.email(), fake.password()
        )
        # Create a couple of person objects
        for _ in range(0, fake.random_int(5, 20)):
            Person.objects.create(
                first_name=fake.first_name(), last_name=fake.last_name()
            )
        self.unique = Person.objects.create(
            first_name=unique_first_name, last_name=unique_last_name
        )
        self.person_count = Person.objects.count()

    def test_person_count(self, setup):
        """Check if the correct number of persons was created"""
        assert Person.objects.count() == self.person_count

    # Initial tests for checking the permissions of the views

    @pytest.mark.parametrize(route_parameters, routes)
    def test_anon_person_views(self, setup, route, permission, status_code, client):
        """Anon user should not be able to access the person views"""
        response = client.get(route)
        assert response.status_code == 302

    @pytest.mark.parametrize(route_parameters, routes)
    def test_admin_person_views(
        self, setup, route, permission, status_code, admin_client
    ):
        """Admin user should be able to access the person list"""
        response = admin_client.get(route)
        assert response.status_code == status_code

    @pytest.mark.parametrize(route_parameters, routes)
    def test_person_user_views(self, setup, route, permission, status_code, client):
        """Normal user should not be able to access the person list"""
        client.force_login(self.user)
        response = client.get(route)
        assert response.status_code == 403

    @pytest.mark.parametrize(route_parameters, routes)
    def test_person_user_with_access(
        self, setup, route, permission, status_code, client
    ):
        """Normal user should be able to access the person views after bein granted permission"""
        content_type = ContentType.objects.get_for_model(Person)
        p = Permission.objects.get(codename=permission, content_type=content_type)
        self.user.user_permissions.add(p)
        client.force_login(self.user)
        response = client.get(route)
        assert response.status_code == status_code

    # More detailed tests checking the content of the responses and sending actual data

    def test_person_list(self, setup, admin_client):
        """Check if the first_name of some person object is listed in the person list"""
        p = Person.objects.order_by("?").first()
        response = admin_client.get("/generic_tests.person/")
        assert p.first_name in response.content.decode()

    def test_person_detail(self, setup, admin_client):
        """Check if the details page of a person lists its details"""
        p = Person.objects.order_by("?").first()
        response = admin_client.get(f"/generic_tests.person/{p.pk}")
        assert p.first_name in response.content.decode()
        assert p.last_name in response.content.decode()

    def test_person_create(self, setup, admin_client):
        """Check if we can create a person using the create route"""
        data = {
            "first_name": fake.first_name(),
            "last_name": fake.last_name(),
        }
        response = admin_client.post("/generic_tests.person/create", data, follow=True)
        assert response.status_code == 200
        assert data["first_name"] in response.content.decode()
        assert data["last_name"] in response.content.decode()
        assert Person.objects.count() == self.person_count + 1

    def test_person_delete(self, setup, admin_client):
        """Check if we can delete a person using the delete route"""
        p = Person.objects.order_by("?").first()
        response = admin_client.get(f"/generic_tests.person/delete/{p.pk}")
        assert response.status_code == 200
        assert "Confirm deletion of:" in response.content.decode()
        assert str(p) in response.content.decode()

    def test_person_delete_confirmation(self, setup, admin_client):
        """Check if we can delete a person using the delete route"""
        p = self.unique
        response = admin_client.post(
            f"/generic_tests.person/delete/{p.pk}", follow=True
        )
        assert response.status_code == 200
        assert p.first_name not in response.content.decode()
        assert p.last_name not in response.content.decode()
        assert Person.objects.count() == self.person_count - 1

    def test_person_delete_redirect(self, setup, admin_client):
        """Check if we can delete a person using the delete route"""
        p = self.unique
        response = admin_client.post(
            f"/generic_tests.person/delete/{p.pk}?redirect=/generic_tests.person/create"
        )
        assert response.status_code == 302
        assert response.headers["Location"] == "/generic_tests.person/create"

    def test_person_update(self, setup, admin_client):
        """Check if we can update a person"""
        p = Person.objects.order_by("?").first()
        data = {
            "first_name": fake.first_name(),
            "last_name": fake.last_name(),
        }
        response = admin_client.post(
            f"/generic_tests.person/update/{p.pk}", data, follow=True
        )
        assert response.status_code == 200
        assert Person.objects.get(pk=p.pk).last_name == data["last_name"]
        assert Person.objects.get(pk=p.pk).first_name == data["first_name"]

    def test_person_autocomplete(self, setup, admin_client):
        """Check if we find a person using autocomplete"""
        p = self.unique
        response = admin_client.get(
            f"/generic_tests.person/autocomplete?q={p.first_name}"
        )
        assert response.status_code == 200
        expected = {
            "results": [{"id": str(p.pk), "selected_text": str(p), "text": str(p)}],
            "pagination": {"more": False},
        }
        assert response.json() == expected

    def test_person_table_export_filename(self, rf, admin_user):
        """Check if the person table provides the correct export filename"""
        view = List(model=Person)
        assert view.get_export_filename("ext") == "table.ext"

    def test_dummy_table_export_filename(self, rf, admin_user):
        """Check if the dummy table provides the correct export filename"""
        view = List(model=Dummy)
        assert view.get_export_filename("ext") == "dummytable.ext"

    def test_dummy_table_class(self, admin_client):
        request = admin_client.get("/generic_tests.dummy/")
        assert isinstance(request.context["table"], DummyTable)



import re

from AcdhArcheAssets.uri_norm_rules import get_rules
from django.contrib.contenttypes.models import ContentType
from django.shortcuts import get_object_or_404
from rdflib import Graph, Literal, Namespace, URIRef
from rdflib.namespace import OWL, RDF, RDFS
from rest_framework.reverse import reverse
from rest_framework.serializers import (
    BaseSerializer,
    CharField,
    HyperlinkedModelSerializer,
    HyperlinkedRelatedField,
    IntegerField,
    Serializer,
    SerializerMethodField,
)

from apis_core.generic.utils.rdf_namespace import APPELLATION, ATTRIBUTES, CRM
from apis_core.utils.settings import rdf_namespace_prefix


class GenericHyperlinkedRelatedField(HyperlinkedRelatedField):
    def get_url(self, obj, view_name, request, format):
        contenttype = ContentType.objects.get_for_model(obj, for_concrete_model=True)
        url_kwargs = {"contenttype": contenttype, "pk": obj.pk}
        return reverse(
            "apis_core:generic:genericmodelapi-detail",
            kwargs=url_kwargs,
            request=request,
            format=format,
        )

    def use_pk_only_optimization(self):
        # We have the complete object instance already. We don't need
        # to run the 'only get the pk for this relationship' code.
        return False


class GenericHyperlinkedIdentityField(GenericHyperlinkedRelatedField):
    def __init__(self, view_name=None, **kwargs):
        assert view_name is not None, "The `view_name` argument is required."
        kwargs["read_only"] = True
        kwargs["source"] = "*"
        super().__init__(view_name, **kwargs)


class GenericHyperlinkedModelSerializer(HyperlinkedModelSerializer):
    serializer_related_field = GenericHyperlinkedRelatedField
    serializer_url_field = GenericHyperlinkedIdentityField


def serializer_factory(
    model,
    serializer=GenericHyperlinkedModelSerializer,
    action="detail",
    fields="__all__",
    **kwargs,
):
    defaultmeta = type(str("Meta"), (object,), {"fields": fields})
    meta = getattr(serializer, "Meta", defaultmeta)
    meta.model = model
    serializer = type(
        str("%s%sModelSerializer" % (model._meta.object_name, action)),
        (serializer,),
        {"Meta": meta},
    )
    return serializer


class ContentTypeInstanceSerializer(Serializer):
    id = IntegerField(required=True)
    content_type = CharField(required=True)

    def to_internal_value(self, data):
        data = super().to_internal_value(data)
        app_label, model = data.get("content_type").split(".")
        content_type = get_object_or_404(ContentType, app_label=app_label, model=model)
        return get_object_or_404(content_type.model_class(), pk=data.get("id"))


class SimpleObjectSerializer(Serializer):
    url = GenericHyperlinkedIdentityField(
        view_name="apis_core:generic:genericmodelapi-detail"
    )
    label = SerializerMethodField()
    content_type_key = SerializerMethodField()
    content_type_name = SerializerMethodField()

    class Meta:
        fields = ["url", "label", "content_type_key", "content_type_name"]

    def get_label(self, obj) -> str:
        return str(obj)

    def get_content_type_key(self, obj) -> str:
        content_type = ContentType.objects.get_for_model(obj)
        return f"{content_type.app_label}.{content_type.model}"

    def get_content_type_name(self, obj) -> str:
        content_type = ContentType.objects.get_for_model(obj)
        return content_type.name


class GenericModelCidocSerializer(BaseSerializer):
    def __init__(self, *args, **kwargs):
        self.rdf_nsp_base = rdf_namespace_prefix()
        self.appellation_nsp_prefix = f"{self.rdf_nsp_base}-appellation"
        self.attr_nsp_prefix = f"{self.rdf_nsp_base}-attr"
        super().__init__(*args, **kwargs)

    def create_sameas(self, g, instance):
        # add the ID as APIS Identifier
        apis_id = URIRef(ATTRIBUTES[f"apis-identifier_{instance.pk}"])
        g.add((apis_id, RDF.type, CRM.E42_Identifier))
        g.add((apis_id, RDFS.label, Literal(instance.pk)))

        # APIS internal identifier type
        apis_id_type = URIRef(ATTRIBUTES["apis-identifier_type"])
        g.add((apis_id, CRM.P2_has_type, apis_id_type))
        g.add((apis_id_type, RDF.type, CRM.E55_Type))
        g.add((apis_id_type, RDFS.label, Literal("APIS internal identifier")))
        g.add((self.instance_uri, CRM.P1_is_identified_by, apis_id))

        for uri in instance.uri_set():
            g.add((self.instance_uri, OWL.sameAs, URIRef(uri.uri)))

            for x in get_rules():
                if m := re.match(x["match"], uri.uri):
                    id_type = URIRef(ATTRIBUTES[x["name"] + "-identifier_type"])
                    g.add((id_type, RDF.type, CRM.E55_Type))
                    g.add((id_type, RDFS.label, Literal(x["name"] + " ID")))
                    id_uri = URIRef(
                        ATTRIBUTES[x["name"] + f"-identifier_{instance.pk}"]
                    )
                    g.add((id_uri, RDF.type, CRM.E42_Identifier))
                    g.add((id_uri, CRM.P2_has_type, id_type))
                    g.add(
                        (
                            self.instance_uri,
                            CRM.P1_is_identified_by,
                            id_uri,
                        )
                    )
                    g.add((id_uri, RDFS.label, Literal(m[1])))

        return g

    def to_representation(self, instance):
        g = Graph()

        g.namespace_manager.bind("crm", CRM, replace=True)
        g.namespace_manager.bind("owl", OWL, replace=True)

        g.namespace_manager.bind(self.appellation_nsp_prefix, APPELLATION, replace=True)
        g.namespace_manager.bind(self.attr_nsp_prefix, ATTRIBUTES, replace=True)

        self.instance_namespace = Namespace(instance.get_namespace_uri())
        g.namespace_manager.bind(
            instance.get_namespace_prefix(), self.instance_namespace
        )

        self.instance_uri = URIRef(self.instance_namespace[str(instance.id)])

        # Add properties
        self.appellation_uri = URIRef(APPELLATION[str(instance.id)])
        g.add(
            (
                self.appellation_uri,
                RDF.type,
                CRM.E33_E41_Linguistic_Appellation,
            )
        )
        g.add(
            (
                self.instance_uri,
                CRM.P1_is_identified_by,
                self.appellation_uri,
            )
        )
        g.add((self.appellation_uri, RDFS.label, Literal(str(instance))))
        g.add((self.instance_uri, RDFS.label, Literal(str(instance))))
        # Add RDF types
        for rdf_type in instance.get_rdf_types():
            g.add((self.instance_uri, RDF.type, rdf_type))
        # Add sameAs links
        g = self.create_sameas(g, instance)

        return g



from django import template
from django.contrib.contenttypes.models import ContentType
from django.db.models import Case, Q, Value, When

from apis_core.generic.helpers import first_member_match, module_paths
from apis_core.relations.models import Relation
from apis_core.relations.tables import RelationsListTable
from apis_core.relations.utils import relation_content_types

register = template.Library()


@register.simple_tag
def possible_relation_types_from(obj) -> list[ContentType]:
    return relation_content_types(any_model=type(obj))


@register.simple_tag
def get_relation_targets_from(obj) -> list[ContentType]:
    relations = relation_content_types(any_model=type(obj))
    types = set()
    for model in [relation.model_class() for relation in relations]:
        if type(obj) in model.obj_list():
            types.update(model.subj_list())
        if type(obj) in model.subj_list():
            types.update(model.obj_list())
    return sorted(
        list(map(ContentType.objects.get_for_model, types)), key=lambda x: x.name
    )


@register.simple_tag
def relations_from(from_obj, relation_type: ContentType = None):
    from_content_type = ContentType.objects.get_for_model(from_obj)
    relation = Relation
    if relation_type is not None:
        relation = relation_type.model_class()

    relations = (
        relation.objects.filter(
            Q(subj_content_type=from_content_type, subj_object_id=from_obj.id)
            | Q(obj_content_type=from_content_type, obj_object_id=from_obj.id)
        )
        .annotate(
            forward=Case(
                When(
                    subj_content_type=from_content_type,
                    subj_object_id=from_obj.id,
                    then=Value(True),
                ),
                default=Value(False),
            )
        )
        .select_subclasses()
    )
    return relations


@register.simple_tag(takes_context=True)
def relations_list_table(context, relations, suffix=None):
    suffixes = ["RelationsTable"]
    if suffix:
        suffixes.insert(0, f"{suffix}RelationsTable")
    table_modules = ()
    for suffix in suffixes:
        table_modules += module_paths(
            type(context["object"]), path="tables", suffix=suffix
        )
    table_class = first_member_match(table_modules, RelationsListTable)
    return table_class(relations, request=context["request"])


@register.simple_tag
def get_relation_content_types():
    return relation_content_types()


@register.simple_tag
def get_relation_between_content_types(source: ContentType, target: ContentType):
    return relation_content_types(
        combination=(source.model_class(), target.model_class())
    )


@register.simple_tag(takes_context=True)
def relations_instances_from_relation_types(context, relation_types):
    relations = []
    for relation_type in relation_types:
        relations.extend(relations_from(context["object"], relation_type))
    return relations



# Generated by Django 5.0.6 on 2024-08-27 08:39

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("relations", "0001_initial"),
        ("sample_project", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="IsCousinOf",
            fields=[
                (
                    "relation_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="relations.relation",
                    ),
                ),
            ],
            bases=("relations.relation",),
        ),
        migrations.CreateModel(
            name="IsPartOf",
            fields=[
                (
                    "relation_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="relations.relation",
                    ),
                ),
            ],
            bases=("relations.relation",),
        ),
        migrations.CreateModel(
            name="IsSiblingOf",
            fields=[
                (
                    "relation_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="relations.relation",
                    ),
                ),
            ],
            bases=("relations.relation",),
        ),
        migrations.CreateModel(
            name="LivesIn",
            fields=[
                (
                    "relation_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="relations.relation",
                    ),
                ),
            ],
            bases=("relations.relation",),
        ),
    ]



from drf_spectacular.types import OpenApiTypes
from drf_spectacular.utils import extend_schema_field
from rest_framework import serializers


class ModelChangeSerializer(serializers.Serializer):
    field = serializers.CharField()
    old = serializers.SerializerMethodField(method_name="get_field_data_old")
    new = serializers.SerializerMethodField(method_name="get_field_data_new")

    def get_data_for_m2m_field(self, value, field):
        sst = [
            str(obj2)
            for obj2 in field.related_model.objects.filter(
                pk__in=[obj3[field.attname] for obj3 in value]
            )
        ]
        return " | ".join(sst)

    def get_field_data(self, obj, new: bool):
        if new:
            value = obj.new
        else:
            value = obj.old
        if obj.field in [x.name for x in self._parent_object._meta.local_fields]:
            field = self._parent_object._meta.get_field(obj.field)
            if field.is_relation and value is not None:
                return str(getattr(self._args[1], obj.field))
            return value
        else:
            for field in self._parent_object._history_m2m_fields:
                if field.attname == obj.field:
                    return self.get_data_for_m2m_field(value, field)
        return obj

    def get_field_data_new(self, obj):
        repr = self.get_field_data(obj, True)
        return repr

    def get_field_data_old(self, obj):
        repr = self.get_field_data(obj, False)
        return repr

    def __init__(self, instance=None, parent_object=None, **kwargs):
        self._parent_object = parent_object
        super().__init__(instance, **kwargs)


class HistoryLogSerializer(serializers.Serializer):
    diff = serializers.SerializerMethodField()
    timestamp = serializers.DateTimeField(source="history_date")
    user = serializers.CharField(source="history_user")
    action = serializers.SerializerMethodField()
    model = serializers.CharField(source="instance.__class__.__name__")
    module = serializers.CharField(source="instance.__class__._meta.app_label")
    instance = serializers.CharField()
    object_id = serializers.IntegerField(source="id")
    history_id = serializers.IntegerField()

    @extend_schema_field(OpenApiTypes.OBJECT)
    def get_diff(self, obj):
        if obj.history_type == "-":
            return None
        changed_fields = []
        changes = []
        for change in obj.get_diff():
            changed_fields.append(change.field)
            changes.append(ModelChangeSerializer(change, obj).data)
        return {"changed_fields": changed_fields, "changes": changes}

    def get_action(self, obj) -> str:
        match obj.history_type:
            case "+":
                return "created"
            case "~":
                return "changed"
            case "-":
                return "deleted"


class HistoryObjectSerializer(serializers.Serializer):
    history = serializers.SerializerMethodField()
    model = serializers.CharField(source="__class__.__name__")
    module = serializers.CharField(source="_meta.app_label")
    instance = serializers.SerializerMethodField()
    object_id = serializers.IntegerField(source="id")

    def get_instance(self, obj) -> str:
        return str(obj)

    @extend_schema_field(HistoryLogSerializer(many=True))
    def get_history(self, obj):
        return HistoryLogSerializer(obj.get_history_data(), many=True).data



from crispy_forms.helper import FormHelper
from crispy_forms.layout import Submit
from django import forms
from django.utils.translation import gettext_lazy as _

from apis_core.collections.models import SkosCollection


class CollectionObjectForm(forms.Form):
    collections = forms.ModelMultipleChoiceField(
        required=False, queryset=SkosCollection.objects.all(), label=_("Collections")
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.add_input(Submit("submit", _("Submit")))
